﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using ThinkGeo.MapSuite.Drawing;
using ThinkGeo.MapSuite.Layers;
using ThinkGeo.MapSuite.Shapes;
using ThinkGeo.MapSuite.Styles;
using ThinkGeo.MapSuite.WinForms;

namespace ScaleLineIssue
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();

            SetupMap();            
        }

        public LayerOverlay CultureOverlay
        {
            get { return (LayerOverlay)winformsMap1.Overlays["CultureOverlay"]; }
        }

        public void SetupMap()
        {
            winformsMap1.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
           
            // This is bad
            Proj4Projection proj = new Proj4Projection(32065, 32065);
            winformsMap1.MapUnit = ThinkGeo.MapSuite.GeographyUnit.Feet;

            // This is good
            //Proj4Projection proj = new Proj4Projection(32615, 32615);
            //winformsMap1.MapUnit = ThinkGeo.MapSuite.GeographyUnit.Meter;

            LayerOverlay culture = new LayerOverlay();
            culture.Name = "CultureOverlay";
            culture.DrawingExceptionMode = DrawingExceptionMode.DrawException;
            winformsMap1.Overlays.Add("CultureOverlay", culture);

            SetScaleLineAdornment(winformsMap1);
            CreateGraticule(proj.ExternalProjectionParametersString);
            AddBox(proj.ExternalProjectionParametersString);
        }

        private void AddBox(string proj)
        {
            InMemoryFeatureLayer imfl = new InMemoryFeatureLayer();
            imfl.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = new AreaStyle(new GeoPen(GeoColor.StandardColors.Red));
            imfl.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            imfl.Name = "Box";

            RectangleShape rect = new RectangleShape(2770000, 10005000, 2775000, 10000000);

            imfl.InternalFeatures.Add(new Feature(rect));

            imfl.FeatureSource.Projection = new Proj4Projection(proj, proj);

            CultureOverlay.Layers.Add("Box", imfl);
            winformsMap1.CurrentExtent = rect;
            winformsMap1.Refresh();
        }

        public void SetScaleLineAdornment(WinformsMap map)
        {
            ScaleLineAdornmentLayer scaleLine = new ScaleLineAdornmentLayer(AdornmentLocation.LowerLeft);
            scaleLine.XOffsetInPixel = 25;
            scaleLine.YOffsetInPixel = -5;

            if (map.AdornmentOverlay.Layers.Contains("ScaleLine")) map.AdornmentOverlay.Layers.Remove("ScaleLine");
            map.AdornmentOverlay.Layers.Add("ScaleLine", scaleLine);
        }

        public void CreateGraticule(string proj)
        {
            XYGraticuleAdornmentLayer grat = new XYGraticuleAdornmentLayer();

            grat.GraticuleLineStyle.OuterPen.Color = GeoColor.FromArgb(125, GeoColor.StandardColors.LightBlue);
            grat.GraticuleTextFont = new GeoFont("Arial", 9, DrawingFontStyles.Bold);
            grat.Projection = new Proj4Projection(proj, proj);

            CultureOverlay.Layers.Add("graticule", grat);
        }
    }

    public class XYGraticuleAdornmentLayer : GraticuleFeatureLayer
    {
        enum LineType { Meridian, Parallel };
        Collection<double> intervals = new Collection<double>();

        #region Properties
        /// <summary>
        /// Number of lines per width of full map
        /// </summary>
        public int GraticuleDensity { get; set; }

        /// <summary>
        /// Set to false to display lat/long labels off map.  Set to 
        /// true to display lat/long labels on map.  Default is true;
        /// </summary>
        public bool DisplayLabelsOnMap { get; set; }

        /// <summary>
        /// Offset, in pixels, from longitude upper point where longitude label will be placed.  Default is 10 if 
        /// DisplayLabelsOnMap is false and 6 is DisplayLabelsOnMap is true.
        /// </summary>
        public float LongitudeLabelOffset { get; set; }

        /// <summary>
        /// Offset from map edge in pixels where longitude label will be placed.  Default is 15 if DisplayLabelsOnMap
        /// is false and 0 if DisplayLabelsOnMap is true.
        /// </summary>
        public float LatitudeLabelOffset { get; set; }

        /// <summary>
        /// Longitude offset from buttom of map.  Valid values should be greater than 0 and is based on 
        /// projection being used by layer.  For example, smallest value for EPSG:4326 is 0.  Values outside of 
        /// 0  to +180 for  EPSG:4326 are considered invalid causing lines to span map.
        /// </summary>
        protected float LongitudeOffset { get; set; }

        /// <summary>
        /// Latitude offset from left edge of map.  Valid values should be greater than 0 and is based on 
        /// projection being used by layer.  For example, smallest value for EPSG:4326 is 0.  Values outside of 
        /// 0 and +360 for EPSG:4326 are considered invalid causing lines to span map.
        /// </summary>
        protected float LatitudeOffset { get; set; }

        /// <summary>
        /// This is max extent based on layer projection.
        /// </summary>
        protected RectangleShape MaxBoundingBox { get; set; }

        /// <summary>
        /// Set to true for XY axis
        /// </summary>
        protected bool IsXyAxis { get; set; }
        #endregion


        public XYGraticuleAdornmentLayer()
            : this(10)
        {
        }

        public XYGraticuleAdornmentLayer(int gridLineDensity)
        {
            GraticuleDensity = gridLineDensity == 0 ? 10 : gridLineDensity;
            SetDefaultIntervals();
            this.DisplayLabelsOnMap = false;

            if (!this.DisplayLabelsOnMap)
            {
                this.LongitudeLabelOffset = 3;
                this.LatitudeLabelOffset = 5;
            }
            else
            {
                this.LongitudeLabelOffset = 2;
                this.LatitudeLabelOffset = 0;
            }

            LongitudeOffset = 180;
            LatitudeOffset = 360;
        }


        #region Protected
        public static XYGraticuleAdornmentLayer GetXYAxisLayer()
        {
            return new XYGraticuleAdornmentLayer()
            {
                LongitudeOffset = 3.5F,
                LatitudeOffset = 4,
                DisplayLabelsOnMap = false,
                IsXyAxis = true,
            };
        }

        protected override void DrawCore(GeoCanvas canvas, Collection<SimpleCandidate> labelsInAllLayers)
        {
            RectangleShape currentExtent = canvas.CurrentWorldExtent;

            if (MaxBoundingBox == null)
            {
                RectangleShape espg4328Extent = new RectangleShape(-180, 90, 180, -90);
                if (this.Projection == null)
                    MaxBoundingBox = espg4328Extent;
                else
                    MaxBoundingBox = this.Projection.ConvertToExternalProjection(espg4328Extent);
            }

            double currentMinX = currentExtent.UpperLeftPoint.X;
            if (currentMinX < MaxBoundingBox.UpperLeftPoint.X)
                currentMinX = MaxBoundingBox.UpperLeftPoint.X;

            // set contraint on x
            double currentMaxX = currentExtent.LowerRightPoint.X;
            if (LatitudeOffset < MaxBoundingBox.Width)
                currentMaxX = currentMinX + LatitudeOffset;
            if (currentMaxX > MaxBoundingBox.UpperRightPoint.X)
                currentMaxX = MaxBoundingBox.LowerRightPoint.X;

            double currentMinY = currentExtent.LowerLeftPoint.Y;
            if (currentMinY < MaxBoundingBox.LowerLeftPoint.Y)
                currentMinY = MaxBoundingBox.LowerLeftPoint.Y;

            // set contraint on y
            double currentMaxY = currentExtent.UpperLeftPoint.Y;
            if (LongitudeOffset < MaxBoundingBox.Height)
                currentMaxY = currentMinY + LongitudeOffset;
            if (currentMaxY > MaxBoundingBox.UpperLeftPoint.Y)
                currentMaxY = MaxBoundingBox.UpperLeftPoint.Y;



            //Gets the increment according to the current extent of the map and the graticule density set 
            //by the GrsaticuleDensity property
            double increment = GetIncrement(currentExtent.Width, GraticuleDensity);
            //increment = 50000;

            double startX = CeilingNumber(currentMinX, increment);
            double startY = CeilingNumber(currentMinY, increment);
            double y = 0;
            double x = 0;
            int index = 0;

            //Loop for displaying the meridians (lines of common longitude).
            for (x = startX; x <= currentExtent.LowerRightPoint.X; x += increment)
            {
                //if ((x <= MaxBoundingBox.LowerRightPoint.X) && (x >= MaxBoundingBox.LowerLeftPoint.X))
                {
                    if (index == 0 && IsXyAxis)
                    {
                        // draw entire first line
                        y = MaxBoundingBox.UpperLeftPoint.Y;

                        // update min and max X
                        var offset = currentMaxX - currentMinX;
                        currentMinX = x;
                        currentMaxX = currentMinX + offset;
                    }
                    else
                        y = currentMaxY;

                    LineShape lineShapeMeridian = new LineShape();
                    //lineShapeMeridian.Vertices.Add(new Vertex(x, startY));
                    lineShapeMeridian.Vertices.Add(new Vertex(x, y));
                    lineShapeMeridian.Vertices.Add(new Vertex(x, currentExtent.UpperRightPoint.Y));
                    canvas.DrawLine(lineShapeMeridian, this.GraticuleLineStyle.OuterPen, DrawingLevel.LevelFour);

                    // create and draw label  
                    var labelY = startY + LongitudeLabelOffset;
                    if (!DisplayLabelsOnMap)
                        labelY = startY - LongitudeLabelOffset;

                    double o = increment / 20;
                    o = (canvas.Height - 5) / canvas.Height * increment / 20;

                    labelY = currentExtent.UpperLeftPoint.Y - o;

                    ScreenPointF meridianLabelPosition = ExtentHelper.ToScreenCoordinate(currentExtent, x, labelY, canvas.Width, canvas.Height);
                    var meridianLabel = new GraticuleLabel(x.ToString(), meridianLabelPosition);
                    var locations = new ScreenPointF[] { new ScreenPointF(meridianLabel.Location.X, meridianLabel.Location.Y) };
                    canvas.DrawText(meridianLabel.Label, this.GraticuleTextFont, this.GraticuleTextBrush,
                        new GeoPen(GeoColor.StandardColors.White, 2), locations, DrawingLevel.LevelFour, 8, 0, 0);

                    index++;
                }
            }

            //Loop for displaying the parallels (lines of common latitude).        
            index = 0;
            for (y = startY; y <= currentExtent.UpperLeftPoint.Y; y += increment)
            {
                //if ((y <= MaxBoundingBox.UpperLeftPoint.Y) && (y >= MaxBoundingBox.LowerLeftPoint.Y))
                {
                    if (index == 0 && IsXyAxis)
                    {
                        var offset = currentMaxY - currentMinY;
                        currentMinY = y;
                        currentMaxY = currentMinY + offset;
                        x = MaxBoundingBox.LowerRightPoint.X;
                    }
                    else
                        x = currentMaxX;

                    LineShape lineShapeParallel = new LineShape();
                    //lineShapeParallel.Vertices.Add(new Vertex(startX, y));
                    lineShapeParallel.Vertices.Add(new Vertex(x, y));
                    lineShapeParallel.Vertices.Add(new Vertex(currentExtent.UpperRightPoint.X, y));
                    canvas.DrawLine(lineShapeParallel, this.GraticuleLineStyle.OuterPen, DrawingLevel.LevelFour);

                    //Gets the label and screen position of each parallel.
                    var labelX = startX - LatitudeLabelOffset;
                    if (DisplayLabelsOnMap)
                        labelX = startX + LatitudeLabelOffset;

                    labelX = currentExtent.LowerLeftPoint.X + 50;

                    ScreenPointF parallelLabelPosition = ExtentHelper.ToScreenCoordinate(currentExtent, labelX, y, canvas.Width, canvas.Height);

                    var parallelGraticuleLabel = new GraticuleLabel(y.ToString(), parallelLabelPosition);
                    var locations = new ScreenPointF[] { new ScreenPointF(parallelGraticuleLabel.Location.X, parallelGraticuleLabel.Location.Y) };

                    canvas.DrawText(parallelGraticuleLabel.Label, this.GraticuleTextFont, new GeoSolidBrush(GeoColor.StandardColors.Navy),
                        new GeoPen(GeoColor.StandardColors.White, 2), locations, DrawingLevel.LevelFour, 8, 0, 90);

                    index++;
                }
            }
        }
        #endregion

        #region Privates
        //The intervals need to be added from the smallest to the largest.
        private void SetDefaultIntervals()
        {
            intervals.Add(5000);
            intervals.Add(10000);
            intervals.Add(20000);
            intervals.Add(50000);
            intervals.Add(100000);
            intervals.Add(200000);
            intervals.Add(500000);
            intervals.Add(1000000);
            intervals.Add(2000000);
            intervals.Add(5000000);
            intervals.Add(10000000);
            intervals.Add(20000000);
            intervals.Add(50000000);
            intervals.Add(100000000);
            intervals.Add(200000000);
            intervals.Add(400000000);
            intervals.Add(500000000);
            //intervals.Add(0.0005);
            //intervals.Add(0.001);
            //intervals.Add(0.002);
            //intervals.Add(0.005);
            //intervals.Add(0.01);
            //intervals.Add(0.02);
            //intervals.Add(0.05);
            //intervals.Add(0.1);
            //intervals.Add(0.2);
            //intervals.Add(0.5);
            //intervals.Add(1);
            //intervals.Add(2);
            //intervals.Add(5);
            //intervals.Add(10);
            //intervals.Add(20);
            //intervals.Add(40);
            //intervals.Add(50);
        }

        //Formats the decimal degree value into Degree Minute and Seconds according to the increment. It also looks
        //if the longitude is East or West and the latitude North or South.
        private string FormatLatLong(double value, LineType lineType, double increment)
        {
            string result = "";
            try
            {
                if (increment >= 1)
                {
                    result = DecimalDegreesHelper.GetDegreesMinutesSecondsStringFromDecimalDegree(Math.Abs(value));
                    result = result.Substring(0, result.Length - 9);
                }
                else if (increment >= 0.1)
                {
                    result = DecimalDegreesHelper.GetDegreesMinutesSecondsStringFromDecimalDegree(Math.Abs(value));
                    result = result.Substring(0, result.Length - 5);
                }
                else if (increment >= 0.01)
                {
                    result = DecimalDegreesHelper.GetDegreesMinutesSecondsStringFromDecimalDegree(Math.Abs(value));
                }
                else
                {
                    result = DecimalDegreesHelper.GetDegreesMinutesSecondsStringFromDecimalDegree(Math.Abs(value), 2);
                }

                if (lineType == LineType.Meridian)
                {
                    if (value > 0) result = result + " E";
                    else if (value < 0) result = result + " W";
                }

                if (lineType == LineType.Parallel)
                {
                    if (value > 0) result = result + " N";
                    else if (value < 0) result = result + " S";
                }
            }
            catch
            { result = "N/A"; }
            finally { }

            return result;
        }

        //Function used for determining the degree value to use according to the interval.
        private double CeilingNumber(double Number, double Interval)
        {
            double result = 0;
            double IEEERemainder = Math.IEEERemainder(Number, Interval);
            if (IEEERemainder > 0)
                result = (Number - IEEERemainder) + Interval;
            else if (IEEERemainder < 0)
                result = Number + Math.Abs(IEEERemainder);
            else
                result = Number;
            return result;
        }

        //Gets the increment to used according to the with of the current extent and the graticule density.
        private double GetIncrement(double CurrentExtentWidth, double Divisor)
        {
            double result = 0;
            double rawInterval = CurrentExtentWidth / Divisor;

            for (int i = 0; i < intervals.Count; i++)
            {
                if (rawInterval < intervals[i])
                {
                    result = intervals[i];
                    break;
                }
            }
            if (result == 0) result = intervals[intervals.Count - 1];
            return result;
        }

        //Structure used for labeling the meridians and parallels at the desired location on the map in screen coordinates.
        private struct GraticuleLabel
        {
            public string Label;
            public ScreenPointF Location;

            public GraticuleLabel(string Label, ScreenPointF Location)
            {
                this.Label = Label;
                this.Location = Location;
            }
        }

        #endregion
    }
}
